#include "parse.h"
#include "log.h"
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

char *get_token_d(struct parse *p) {
	static char ret[1024];
	struct parse *q = p;
	strcpy(ret,"");
	while(q) {
		if(q->data) {
			if(!strcmp(ret,"")) strcpy(ret,q->data);
			else strcat(ret,q->data);
		}

		q = q->next;
	}

	return(ret);
}
	
char *get_token_s(struct parse *parse, int num) {
	struct parse *p;
	p = get_token(parse,num);
	if(p)
		return(p->data);
	else	
		return(NULL);
}

char *get_data(struct parse_hdr *parse, char *name) {
	struct parse_hdr *p;
	struct parse *p2;
	p = get_entry(parse,name);
	if(!p) return(NULL);
	p2 = get_token(p->data,1);
	return(p2->data);
}

int get_data_int(struct parse_hdr *parse, char *name) {
	struct parse_hdr *p;
	struct parse *p2;
	p = get_entry(parse,name);
	if(!p) return(-1);
	p2 = get_token(p->data,1);
	if(!p2 || !p2->data) return(-1);
	return(atoi(p2->data));
}

char *get_one_data(struct parse_hdr *parse, char *name) {
	struct parse_hdr *p;
	struct parse *p2;
	p = get_one(parse,name);
	if(!p) return(NULL);
	p2 = get_token(p->data,1);
	return(p2->data);
}

int get_one_data_int(struct parse_hdr *parse, char *name) {
	struct parse_hdr *p;
	struct parse *p2;
	p = get_one(parse,name);
	if(!p) return(-1);
	p2 = get_token(p->data,1);
	if(!p2 || !p2->data) return(-1);
	return(atoi(p2->data));
}
	
struct parse *get_token(struct parse *parse, int num) {
	/* get token */
	struct parse *parsed = parse;
	int i = num;

	while(parsed && (i != 0)) {
		parsed = parsed->next;
		i--;
	}

	return(parsed);
}

struct parse_hdr *get_entry(struct parse_hdr *parse, char *name) {
	/* get entry with token 0 */
	struct parse_hdr *parse_hdr = parse;
	while(parse_hdr) {
		if(!strcmp(parse_hdr->data->data,name)) 	
			return(parse_hdr);
		parse_hdr = parse_hdr->next;
	}
	return(NULL);
}

struct parse_hdr *get_one(struct parse_hdr *parse, char *name) {
	struct parse_hdr *get;
	get = get_entry(parse,name);
	if(!get) return(NULL);
	if(get_entry(get->next,name))
		return(NULL);
	return(get);
}

struct parse_hdr *parse_lines(char *lines) {
	char workbuf[MAX_PARSE];
	int i=0;
	int len=0;
	struct parse_hdr *hdr = NULL;
	struct parse_hdr *head = NULL;

	while(i<=(strlen(lines)+1)) {
		switch(lines[i]) {
			case'\0':
			case '\r':
			case '\n':
				workbuf[len] = '\0';
				if(len) {
					if(hdr) {
						if(!(hdr->next = (struct parse_hdr *) malloc(sizeof(struct parse_hdr)))) 
							return(NULL);
						hdr = hdr->next;
					} else {
						if(!(hdr = (struct parse_hdr *) malloc(sizeof(struct parse_hdr))))
							return(NULL);
						head = hdr;
					}
					hdr->next = NULL;
					hdr->data = parse_line(workbuf);
				}
				len = 0;
				break;
			default:
				workbuf[len++] = lines[i];
		}
	i++;
	}
	
	return(head);
}
			
struct parse *parse_line(char *line) {
	int i=0;
	int len=0;
	int quote = 0;
	int comment = 0;
	int process = 0;
	int stop = 0;
	char workbuf[MAX_PARSE];
	struct parse *parse = NULL;
	struct parse *parse_head = NULL;
	char c;

	if(!line) return(NULL);

	while(!stop) {
		c = line[i];
		switch(c) {
			case '"':
				if(!comment) {
					if(quote)  quote = 0;  else  quote = 1; 
				}
				break;
			case ';':
				if(!quote) comment = 1;
			case '\r':
			case '\n':
			case '\0':
				workbuf[len] = '\0';
				process = 1;
				stop = 1;
				break;
			case '\t':
			case ' ':
				if(!comment) {
					if(quote) {
						workbuf[len++] = c;
					} else {
						if(len) {
							workbuf[len] = '\0';
							process = 1;
							len = 0;
						}
					}
				}
				break;
			default:
				if(!comment) 
					workbuf[len++] = c;
		}
		
		if(process) {
			process = 0;
			if(parse) {
				if(!(parse->next = (struct parse *) malloc(sizeof(struct parse)))) {
					elog("Can't parse line - malloc error.");
					return(NULL); /* memory leak */
				}
				parse = parse->next;
			} else {
				if(!(parse = (struct parse *) malloc(sizeof(struct parse)))) {
					elog("Can't parse line - malloc error.");
					return(NULL);
				}
				parse_head = parse;
			}
		parse->next = NULL;
		parse->data = strdup(workbuf);
		}
		i++;
	}
	return(parse_head);

}

void ins_one(struct parse_hdr *p, char *data) {
	/* insert a row */
	struct parse_hdr *q;
	struct parse *parse;
	q = p;
	while(q->next) q = q->next;

	if(!(q->next = (struct parse_hdr *) malloc(sizeof(struct parse_hdr)))) {
		elog("ins_one(): Unable to malloc()");
		return;
	}

	parse = parse_line(data);
	q = q->next;
	q->next = NULL;
	q->data = parse;
	return;
}

struct parse_hdr *parse_file(char *filename) {
	struct parse_hdr *parse_hdr_head = NULL;
	struct parse_hdr *parse_hdr;
	struct parse *parse;
	char inbuf[MAX_PARSE];
	FILE *fp;

	if(!filename) {
		elog("no filename given to parse_file");
		return(NULL);
	}

	if(!(fp = fopen(filename,"ro"))) {
		elog("File not found at parse_file: %s",filename);
		return(NULL);
	}

	while(fgets(inbuf,MAX_PARSE,fp)) {
		parse = parse_line(inbuf);
		if(parse) {
			if(parse_hdr_head) {
				if(!(parse_hdr->next = (struct parse_hdr *) malloc(sizeof(struct parse_hdr)))) {
					elog("Couldn't parse_file: malloc error");
					return(NULL);
				}
			parse_hdr = parse_hdr->next;
			} else {
				if(!(parse_hdr_head = (struct parse_hdr *) malloc(sizeof(struct parse_hdr)))) {
					elog("couldn't parse_file: malloc error");
					return(NULL);
				}
			parse_hdr = parse_hdr_head;
			}
		
			parse_hdr->next = NULL;
			parse_hdr->data = parse;
		}
	} /* end while */
	return(parse_hdr_head);
}


void free_parse(struct parse *p) {
	struct parse *p2;

	while(p) {
		p2 = p;
		p = p->next;
		if(p2->data) free(p2->data);
		free(p2);
	}
}

void free_parse_hdr(struct parse_hdr *p) {
	struct parse_hdr *p2;

	while(p) {
		free_parse(p->data);
		p2 = p;
		p = p->next;
		free(p2);
	}
}


