#include "socket.h"
#include "filed.h"
#include "externs.h"
#include "queue.h"
#include "log.h"
#include "main.h"
#include "parse.h"
#include "io.h"

#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <malloc.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <time.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>

void bind_socket(void) {
	struct protoent *proto;
	struct sockaddr_in addr;
	int flag;

	proto = getprotobyname("tcp");

	if((filed->main_fd = socket(PF_INET,SOCK_STREAM,proto->p_proto)) == -1) {
		elog("bind_socket(): couldn't create main socket");
		die();
	}

	flag = 1;
	
	if(setsockopt(filed->main_fd,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(flag))) {
		elog("bind_socket(): couldn't setsockopt");
		die();
	}

	bzero(&addr,sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(get_one_data_int(filed->cf,"port"));
	if(bind(filed->main_fd,&addr,sizeof(addr))) {
		elog("bind_socket(): couldn't bind main socket");
		die();
	}

	if(listen(filed->main_fd,5)) {
		elog("bind_socket(): couldn't set up listen on main socket");
		die();
	}

	if((fcntl(filed->main_fd,F_SETFL,FNDELAY))) {
		elog("bind_socket(): couldn't set main socket non blocking");
		die();
	}

	return;
}

void handle_connections(void) {
	fd_set wrfs;
	fd_set rdfs;
	fd_set errfs;
	struct timeval timeout;
	struct connection *c;
	int cap = (filed->main_fd + 1);

	timeout.tv_sec = 0;
	timeout.tv_usec = 1;

	FD_ZERO(&rdfs);
	FD_ZERO(&wrfs);
	FD_ZERO(&errfs);

	FD_SET(filed->main_fd,&rdfs);

	c = filed->con_head;

	while(c) {
		if((c->fd + 1) > cap) 
			cap = (c->fd + 1);
		FD_SET(c->fd,&rdfs);
		FD_SET(c->fd,&errfs);
		if(pending_tx(c))
			FD_SET(c->fd,&wrfs);
	
		c = c->next;
	}

	select(cap,&rdfs,&wrfs,&errfs,&timeout);

	c = filed->con_head;

	while(c) {
		if(c->flags & CFLAG_GONE_AWAY) {
			c = c->next;
			continue;
		}
		if(FD_ISSET(c->fd,&wrfs)) 
			write_connection(c);
		if(FD_ISSET(c->fd,&rdfs)) 
			read_connection(c);
		if(FD_ISSET(c->fd,&errfs))
 			err_connection(c); 
 
		c = c->next;
	}


	if(FD_ISSET(filed->main_fd,&rdfs))
		accept_connection();
	
	return;			
		
}

void write_connection(struct connection *c) {
	/* handle pending output */
	struct queue *q;
	int i;
	int j;
	int ret;
	int buf;
	int blen;

	i = 0;
	blen = sizeof(buf);

	/* we only permit output of three blocks at once */

	while((q = c->out) && i++ < 3) {
		if(q->flags & QFLAG_INPROGRESS) {
			ret = getsockopt(c->fd,SOL_SOCKET,SO_ERROR,&buf,&blen);
			if(ret || buf) {
				if(ret) errno = ret; else errno = buf;
				if(!(c->flags & CFLAG_GONE_AWAY)) c->flags += CFLAG_GONE_AWAY;
				dlog(DEBUG_CON,"[%d] connect to %s failed",c->id,c->remote_addr);
				c_disconnect(c);
				return;
			} 
		
			c->out = rm_queue(c->out,q);
			continue;
		}
		
		/* try a write */
		j = send(c->fd,q->data,q->len,MSG_NOSIGNAL);
		if(j < q->len) {
			q->offset = j;
			break;
		} else {
			c->last_msg = time(NULL);
		}
	
	c->out = rm_queue(c->out,q);
	}

	return;
}

void read_connection(struct connection *c) {
	/* handle pending input */
	int j;
	int k;
	int l;
	char inbuf[1024];
	char outbuf[1024];
	struct queue *q;
	
	/* we assume that reads will stay atomic */

	dlog(DEBUG_CODE,"read_connection() called");	

	j = recv(c->fd,inbuf,1024,MSG_NOSIGNAL);
	inbuf[j] = '\000';

	if((j==0) || (j == -1)) {
		dlog(DEBUG_CODE,"read_connection(): host has gone away");	
		c_hangup(c);
		if(!(c->flags & CFLAG_GONE_AWAY)) c->flags += CFLAG_GONE_AWAY;
		return;
	}

	dlog(DEBUG_IO,"read_connection(): %d read [%s]",c->id,inbuf);

	/* check to see if tail end has offset */

	q = get_tail(c->in);

	c->last_recv = time(NULL);
	
	if(q) {
		if((l = q->offset)) {
			strncpy(outbuf,q->data,q->len);
		}	
	} else l = 0;

	for(k=0;k<j;k++) {
		switch(inbuf[k]) {
			case '\r':
			case '\n':
			case '\000':
				if(l == 0) break;
				outbuf[l] = '\000';
				if(q && q->offset) {
					free(q->data);
					q->data = strdup(outbuf);
					q->len = l;
					q->offset = 0;
					dlog(DEBUG_CODE,"read_connection(): updated queue segment: [%s]",outbuf);
				} else {
					c->in = ins_queue(c->in,outbuf,0);
					dlog(DEBUG_CODE,"read_connection(): inserted queue segment: [%s]",outbuf);
				}
				l = 0;
				break;
			default:
				outbuf[l++] = inbuf[k];
				break;
		}
	}

	if(l) {
		dlog(DEBUG_CODE,"read_connection() ended with offset %d/%d on [%s]",l,k,outbuf);
		q = get_tail(c->in);
		if(q) q->offset = l;
	}
	
	return;

}

void err_connection(struct connection *c) {
	/* handle pending errors */
	dlog(DEBUG_CON,"err_connection() called on socket %d",c->fd);
	/* not really sure what will cause this yet */
}

void accept_connection(void) {
	/* accept a new connection on the main fd */
	int newfd;
	struct sockaddr_in addr;
	struct connection *c;
	char c_addr[128];
	int len;
	

	dlog(DEBUG_CODE,"accept_connection() called");

	len = sizeof(struct sockaddr_in);
	
	bzero(&addr,sizeof(struct sockaddr_in));
	if((newfd = accept(filed->main_fd,(struct sockaddr *) &addr, &len)) == -1) {
		elog("accept_connection(): failed");
		return;
	}

	if(fcntl(newfd,F_SETFL,FNDELAY)) {
		elog("accept_connection() failed to set socket non blocking");
		close(newfd);
		return;
	}
	
	strcpy(c_addr,inet_ntoa(addr.sin_addr));

	if(!(c = new_connection(newfd,c_addr))) {
		elog("accept_connection failed to create a new connection object");
		close(newfd);
	}

	dlog(DEBUG_CON,"New connection %d from %s",c->id,c->remote_addr);
	
	strcpy(c->name,c->remote_addr);
	insert_connection(c);
	return;	
}

struct connection *connect_to(char *ip, int port) {
	/* this attempts to perform a connection - returns NULL if fails */
	struct sockaddr_in addr;
	struct in_addr inaddr;
	struct protoent *proto;
	int newfd;
	struct connection *c;
	int flag = 1;

	if(!(proto = getprotobyname("tcp")))
		return(NULL);
	
	bzero(&addr,sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	if(!(inet_aton(ip,&inaddr))) {
		dlog(DEBUG_CON,"connect_to(): target called [%s] is invalid",ip);
		return(NULL);
	}

	addr.sin_addr = inaddr;

	if((newfd = socket(PF_INET,SOCK_STREAM,proto->p_proto)) == -1) {
		elog("connect_to(): unable to create socket");
		return(NULL);
	}

	/* put the socket non blocking */
	if(setsockopt(newfd,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(flag))) {
		dlog(DEBUG_CON,"connect_to(): setsockopt failed");
		close(newfd);
		return(NULL);
	}

	if(fcntl(newfd,F_SETFL,FNDELAY)) {
		dlog(DEBUG_CON,"connect_to(): fcntl failed");
		close(newfd);
		return(NULL);
	}

	if(connect(newfd,&addr,sizeof(addr)) == -1) {
	
		/*
		 * this gets sneaky
		 * we already know the connection is going to fail
		 * with EINPROGRESS
		 * if it doesn't fail with this, then we return
		 * NULL. If it does fail with this, we create the opject
		 * and set up all the important peices 
		 * and insert into the write queue a INPROCESS flag
		 * when the write function sees this flag,
		 * instead of writing to the connection it will
		 * use getsockopt() to check and see if the connection
		 * is present or not
		 * if not, it will insert on the error queue
		 * a flag to indicate to write
		 * that it should check the status
		 */

		if(errno == EINPROGRESS) {
			c = new_connection(newfd,ip);
			c->out = ins_queue(c->out,NULL,QFLAG_INPROGRESS);
			c->flags = CFLAG_NEWTX;
			dlog(DEBUG_CON,"connection in progress");
			return(c);	
		} else {
			close(newfd);
			return(NULL);
		}
	} else {
		dlog(DEBUG_CON,"connection complete");
		c = new_connection(newfd,ip);
		c->flags = CFLAG_NEWTX;
		return(c);	
	}
	
}	


struct connection *new_connection(int fd, char *addr) {
	/* addr feild is on theory we always know it anyway */

	struct connection *ret;

	if(!(ret = (struct connection *) malloc(sizeof(struct connection)))) {
		elog("new_connection() failed to malloc memory");
		return(NULL);
	}

	ret->flags = CFLAG_NEW;
	ret->type = 0;
	ret->id = get_id();
	if(addr) strcpy(ret->remote_addr,addr);
	else strcpy(ret->remote_addr,"");
	strcpy(ret->name,"");
	ret->next = NULL;
	ret->prev = NULL;
	ret->in = NULL;
	ret->out = NULL;
	ret->err = NULL;
	ret->fd = fd;
	ret->last_msg = time(NULL);
	ret->first_msg = ret->last_msg;
	ret->debug_mask = 0;
	return(ret);
}

void insert_connection(struct connection *c) {
	struct connection *d = filed->con_head;

	if(!d) {
		filed->con_head = c;
		return;
	}

	while(d->next) d = d->next;
	
	d->next = c;
	c->prev = d;
	
	return;
}

void free_connection(struct connection *c) {
	free_queue(c->in);
	free_queue(c->out);
	free_queue(c->err);
	free(c);
	return;
}


void remove_connection(struct connection *d) {
	
	if(!d) return;
	if(d == filed->con_head) {
		filed->con_head = d->next;
	}

	if(d->next) d->next->prev = d->prev;
	if(d->prev) d->prev->next = d->next;
	return;
}

int pending_tx(struct connection *c) {
	if(c && c->out) return(1);
	return(0);
}

int pending_rx(struct connection *c) {
	if(c && c->in) return(1);
	return(0);
}

int gone_away(struct connection *c) {
	if(c && (c->flags & CFLAG_GONE_AWAY)) return(1);
	return(0);
}


void c_send(struct connection *c, char *data,...) {
	char outbuf[1024];
	va_list ap;
	va_start(ap,data);
	vsprintf(outbuf,data,ap);
	va_end(ap);
	strcat(outbuf,"\n");
	
	c->out = ins_queue(c->out,outbuf,0);
}

void c_hangup(struct connection *c) {
	if(c->err && c->err->flags == QFLAG_HANGUP) return;
	c->err = ins_queue(c->err,NULL,QFLAG_HANGUP);
}

void c_disconnect(struct connection *c) {
	if(c->err && c->err->flags == QFLAG_DISCONNECT) return;
	c->err = ins_queue(c->err,NULL,QFLAG_DISCONNECT);
}

int c_send_to(int id, char *data, ...) {
	char outbuf[1024];
	va_list ap;
	struct connection *c = filed->con_head;

	va_start(ap,data);
	vsprintf(outbuf,data,ap);
	va_end(ap);
	strcat(outbuf,"\n");

	while(c) {
		if(c->id == id) {
			c->out = ins_queue(c->out,outbuf,0);
			return(1);
		}
	c = c->next;
	}

	return(0);
}

struct connection *get_connection_by_id(int id) {
	struct connection *r;
	r = filed->con_head;

	while(r) {
		if(r->id == id) return(r);
		r = r->next;
	}
	return(NULL);
}

int get_connection_id_by_name(char *name) {
	struct connection *r = filed->con_head;

	while(r) {
		if(r->name)
		if(!strcasecmp(r->name,name)) 
			return(r->id);
	r = r->next;
	}
	
	return(0);
}

struct connection *get_connection_by_name(char *name) {
	return(get_connection_by_id(get_connection_id_by_name(name)));
}

int check_access(char *addr) {
	struct parse_hdr *ph;
	char *d;
	char *e;

	ph = filed->cf;

	while((ph = get_entry(ph,"allow"))) {
		d = get_token_s(ph->data,1);
		if(!strncmp(addr,d,strlen(d))) {
			e = get_token_s(ph->data,2);
			if(!e) return(1);
			return(atoi(e));	
		}
	ph = ph->next;
	}
	
	return(0);
}

char *c_str_flags(int flags) {
	static char ret[1024];
	strcpy(ret,"");
	if(flags & CFLAG_NEW) strcat(ret,"NEW ");
	if(flags & CFLAG_GONE_AWAY) strcat(ret, "GONE_AWAY ");
	if(flags & CFLAG_NEWTX) strcat(ret,"NEWTX ");
	if(flags & CFLAG_NODATA) strcat(ret,"NODATA ");
	if(flags & CFLAG_IDENTIFIED) strcat(ret,"IDENTIFIED ");
	return(ret);
}

char *c_str_type(int type) {
	switch(type) {
		case CTYPE_HUMAN:
			return("Human");
		case CTYPE_JAVA:
			return("Java");
		case CTYPE_PEER:
			return("Peer");
	}
	return("Unknown");
}

